﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace CommunityDetection
{
	class DecisionTree : Classifier
	{
		DecisionNode root;
		DecisionTreeBuilder treeBuilder;

		public DecisionTree(DecisionTreeBuilder treeBuilder)
		{
			this.treeBuilder = treeBuilder;
		}

		public void create(HashSet<int> trueRows, HashSet<int> rows)
		{
			//create root node
			root = new DecisionNode(this, null, treeBuilder.data, treeBuilder.columnSum);

			//create list of all attributes
			List<int> attributes = new List<int>(treeBuilder.attributeList);

			//splitting
			root.split(attributes, rows, trueRows);

			//calculate critical aplha value
			root.calculateAlpha();
		}

		public List<DecisionNode> bfs(bool internalOnly)
		{
			Queue<DecisionNode> queue = new Queue<DecisionNode>();
			List<DecisionNode> list = new List<DecisionNode>();
			queue.Enqueue(root);
			while (queue.Count > 0)
			{
				DecisionNode node = queue.Dequeue();
				if (!node.isLeaf)
				{
					queue.Enqueue(node.leftNode);
					queue.Enqueue(node.rightNode);
				}
				if (!(internalOnly && node.isLeaf))
					list.Add(node);
			}
			return list;
		}

		public List<DecisionNode> dfs(bool internalOnly)
		{
			Stack<DecisionNode> stack = new Stack<DecisionNode>();
			List<DecisionNode> list = new List<DecisionNode>();
			stack.Push(root);
			while (stack.Count > 0)
			{
				DecisionNode node = stack.Pop();
				if (!node.isLeaf)
				{
					stack.Push(node.leftNode);
					stack.Push(node.rightNode);
				}
				if (!(internalOnly && node.isLeaf))
					list.Add(node);
			}
			return list;
		}

		public class AlphaSort : IComparer<DecisionNode>
		{
			public int Compare(DecisionNode n1, DecisionNode n2)
			{
				if (n1.alpha > n2.alpha)
					return 1;
				else if (n2.alpha > n1.alpha)
					return -1;
				return 0;
			}
		}


		public KeyValuePair<double,DecisionTree> prune(double alpha)
		{
			//copy tree
			DecisionTree tree = new DecisionTree(treeBuilder);

			//copy nodes in tree
			tree.root = new DecisionNode(tree, root, null);

			//get nodes and sort by alpha value
			List<DecisionNode> nodes = tree.dfs(true);

			//prune on lowest alpha
			if (nodes.Count > 0)
			{
				foreach(DecisionNode n in nodes)
				{
					if (n.alpha == alpha)
					{
						n.splitAttribute = -1;
						n.numLeaves = 1;
						n.imp = treeBuilder.measure.impurity(n.numT, n.numF);
						n.leftNode.parent = null;
						n.rightNode.parent = null;
						n.leftNode = null;
						n.rightNode = null;
						n.isLeaf = true;
					}
				}
				tree.getRoot().calculateAlpha();
				return new KeyValuePair<double, DecisionTree>(alpha, tree);
			}
			return new KeyValuePair<double,DecisionTree>(0,null);
		}

		public KeyValuePair<double,DecisionTree> prune()
		{
			//copy tree
			DecisionTree tree = new DecisionTree(treeBuilder);

			//copy nodes in tree
			tree.root = new DecisionNode(tree, root, null);

			//get nodes and sort by alpha value
			List<DecisionNode> nodes = tree.dfs(true);
			nodes.Sort(new AlphaSort());

			//prune on lowest alpha
			if (nodes.Count > 0)
			{
				double alpha = nodes[0].alpha;
				foreach(DecisionNode n in nodes)
				{
					if (n.alpha <= alpha)
					{
						n.splitAttribute = -1;
						n.numLeaves = 1;
						n.imp = treeBuilder.measure.impurity(n.numT, n.numF);
						n.leftNode.parent = null;
						n.rightNode.parent = null;
						n.leftNode = null;
						n.rightNode = null;
						n.isLeaf = true;
					}
				}
				tree.getRoot().calculateAlpha();
				return new KeyValuePair<double, DecisionTree>(alpha, tree);
			}
			return new KeyValuePair<double,DecisionTree>(0,null);
		}

		public DecisionNode getRoot()
		{
			return root;
		}


		public void Write(string file)
		{
			StreamWriter writer = new StreamWriter(file);
			root.print(writer);
			writer.Close();
		}

		public override bool classify(BitArray values, int row)
		{
			return root.classify(values);
		}


		public ImpurityMeasure Measure
		{
			get
			{
				return treeBuilder.getMeasure();
			}
		}
	}
}
