﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace CommunityDetection
{
	public class EmDmRunV2
	{
		string logdir;
		ThreadSafeQueue<Community> queue;
		BeamBuilder builder;
		UndirectedGraph graph;
		BinaryMatrix dm;
		int maxEmDmIterations;
		FinalProgramV2 fp;
		Dictionary<CommunityStamp, NBClassifier> classifiers;
		Dictionary<CommunityStamp, CommunityStamp> classifiedTo;
		ThreadSafeQueue<CommunityStamp> emOnlyList;


		public EmDmRunV2(FinalProgramV2 fp, BeamBuilder builder, ThreadSafeQueue<Community> queue, UndirectedGraph graph, BinaryMatrix dm, string logdir, int maxEmDmIterations, Dictionary<CommunityStamp, NBClassifier> classifiers, Dictionary<CommunityStamp, CommunityStamp> classifiedTo, ThreadSafeQueue<CommunityStamp> emOnlyList)
		{
			this.fp = fp;
			this.builder = builder;
			this.queue = queue;
			this.logdir = logdir;
			this.dm = dm;
			this.graph = graph;
			this.maxEmDmIterations = maxEmDmIterations;
			this.classifiers = classifiers;
			this.classifiedTo = classifiedTo;
			this.emOnlyList = emOnlyList;
		}

		public void explog(string msg)
		{
			fp.explog(msg);
		}

		public void run()
		{
			while (!queue.Empty())
			{
				try
				{
					Community c = queue.Dequeue();
					Console.WriteLine("Starting with new candidate, "+queue.Count+" candidates left");
					processCandidate(c);
				}
				catch (InvalidOperationException)
				{
					break;
				}
			}
		}

		public void writeCommunity(Community c, int iteration, string place, string nextAction, NBClassifier cfr)
		{
			explog(c.ID + ";" + iteration + ";" + place + ";" + c.Size + ";" + c.communityGain() + ";" + c.intraClusterDensity() + ";"+c.modularity()+";"+c.conductance()+";"+c.inverseConductance()+";"+ (cfr == null?"-;-;-;-;-":cfr.getInformation())+ ";" + nextAction + ";" + c.ToString());
		}

		public void processCandidate(Community c)
		{
			StreamWriter writer = new StreamWriter(logdir + "/" + c.ID + ".csv");
			writer.WriteLine("Community: " + c.ID);
			writer.WriteLine();
			int i = 0;
			NBClassifier nbc = null;
			while (maxEmDmIterations == 0 || i<maxEmDmIterations)
			{
				writer.WriteLine("---------------");
				writer.WriteLine("Iteration:;" + i);
				
				//em
				HashSet<int> before = new HashSet<int>(c.getVertices().Keys);
				writer.WriteLine("Phase:;EM");
				writer.WriteLine("Steps:");
				writer.WriteLine("Step;Size;ICD;CG;Vertices");
				writer.WriteLine("0;"+c.Size + ";" + c.intraClusterDensity() + ";" + c.communityGain() + ";" + c.ToString());
				int steps = 1;
				while (c.improve())
				{
					writer.WriteLine(steps+";" + c.Size + ";" + c.intraClusterDensity() + ";" + c.communityGain() + ";" + c.ToString());
					steps++;
				}

				//write information
				CommunityStamp stamp = c.Stamp;

				bool contains;
				lock(classifiers)
					contains = classifiers.ContainsKey(c.Stamp);
				if (contains)
				{
					writeCommunity(c, i, "0. After EM", "Stop, classified earlier ", null);
					lock (classifiers)
					{
						nbc = classifiers[stamp];
					}
					for (int v = 0; v < graph.Count; v++)
					{
						if (nbc.classify(dm.getRow(v)))
							c.AddVertice(v);
						else
							c.RemoveVertice(v);
					}
					lock(classifiedTo)
						classifiedTo[stamp] = c.Stamp;
					writeCommunity(c, i, "1. After stopping", "nothing, finished ", nbc);
					writer.WriteLine();
					writer.WriteLine("Resulted in:");
					writer.WriteLine(";Nodes:;" + c.ToString());
					writer.WriteLine(";CG:;" + c.communityGain());
					writer.WriteLine(";ICD:;" + c.intraClusterDensity());
					writer.WriteLine(";Moldularity:;" + c.modularity());
					writer.WriteLine(";Conductance:;" + c.conductance());
					writer.WriteLine(";InvConductance:;" + c.inverseConductance());
					writer.WriteLine(";Size:;" + c.Size);
					writer.WriteLine("Patterns:"); 
					writer.WriteLine(nbc);
					writer.Close();
					return;
				}
	
				writeCommunity(c, i, "0. After EM", "DM step",null);
				writer.WriteLine();
				writer.WriteLine("Summary:");
				HashSet<int> between = new HashSet<int>(c.getVertices().Keys);
				HashSet<int> added = new HashSet<int>(between.Except(before));
				HashSet<int> removed = new HashSet<int>(before.Except(between));
				writer.WriteLine(";Added:;" + added.Count);
				writer.WriteLine(";Removed:;" + removed.Count);
				writer.WriteLine();
				writer.WriteLine("Community:");
				writer.WriteLine(";Size:" + c.Size);
				writer.WriteLine(";CG:" + c.communityGain());
				writer.WriteLine(";ICD:" + c.intraClusterDensity());
				writer.WriteLine(";Modularity:;"+c.modularity());
				writer.WriteLine(";Conductance:;" + c.conductance());
				writer.WriteLine(";InvConductance:;" + c.inverseConductance());


				//dm
				if (builder != null)
				{
					writer.WriteLine();
					writer.WriteLine();
					writer.WriteLine("Phase:;DM");

					stamp = c.Stamp;

					bool build;
					lock (classifiers)
					{
						build = !classifiers.ContainsKey(stamp);
					}
					if (build)
					{
						BeamClassifier classifier = (BeamClassifier)builder.build(between);
						lock (classifiers)
							classifiers[stamp] = classifier.rebuild();
					}
					lock (classifiers)
						nbc = classifiers[stamp];


					writer.WriteLine();
					writer.WriteLine("Patterns:");
					writer.WriteLine(nbc);
					int[] qualityData = nbc.quality(dm, between);
					writer.WriteLine();
					writer.WriteLine("Summary:");
					writer.WriteLine(";Precision:;" + nbc.precision(qualityData));
					writer.WriteLine(";Recall:;" + nbc.recall(qualityData));
					writer.WriteLine(";TP:;" + qualityData[0]);
					writer.WriteLine(";FP:;" + qualityData[1]);
					writer.WriteLine(";TN:;" + qualityData[2]);
					writer.WriteLine(";FN:;" + qualityData[3]);

					for (int v = 0; v < graph.Count; v++)
					{
						if (nbc.classify(dm.getRow(v)))
							c.AddVertice(v);
						else
							c.RemoveVertice(v);
					}
					lock (classifiedTo)
						classifiedTo[stamp] = c.Stamp;
					HashSet<int> after = new HashSet<int>(c.getVertices().Keys);
					added = new HashSet<int>(after.Except(between));
					removed = new HashSet<int>(between.Except(after));
					writer.WriteLine("Classifier:");
					writer.WriteLine(";Added: " + added.Count);
					writer.WriteLine(";Removed: " + removed.Count);
					writer.WriteLine();
					writer.WriteLine("Community:");
					writer.WriteLine(";Nodes:;" + c.ToString());
					writer.WriteLine(";Size: " + c.Size);
					writer.WriteLine(";CG: " + c.communityGain());
					writer.WriteLine(";ICD: " + c.intraClusterDensity());
					writer.WriteLine(";Modularity: " + c.modularity());
					writer.WriteLine(";Conductance:;" + c.conductance());
					writer.WriteLine(";InvConductance:;" + c.inverseConductance());
					writeCommunity(c, i, "1.After DM", i < maxEmDmIterations || maxEmDmIterations == 0 ? "EM step" : "Stop, max EMDM iterations reached", nbc);
				}
				else
				{
					emOnlyList.Enqueue(c.Stamp);
				}
				writer.WriteLine();
				i++;
			}
			writer.Close();
			writeCommunity(c, i, "2. After procession candidates","Nothing, max EMDM iterations reached", nbc);
		}



	}

	
}
