﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.IO;
using System.Globalization;

namespace CommunityDetection
{
	public class TestClassifierJob
	{
		public double alpha;
		public int maxNumPatterns;
		public int beamWidth;
		public int maxPatternLength;

		public TestClassifierJob(int maxNumPatterns, int beamWidth, int maxPatternLength, double alpha)
		{
			this.maxNumPatterns = maxNumPatterns;
			this.beamWidth = beamWidth;
			this.maxPatternLength = maxPatternLength;
			this.alpha = alpha;
		}

		public override string ToString()
		{
			return maxNumPatterns+", "+beamWidth+", "+maxPatternLength+", "+alpha;
		}
	}

	public class TestClassifier : ExpProgram
	{
		public CandidateFileGenerator generator;
		BinaryMatrix matrix;
		string datafile;
		ThreadSafeQueue<TestClassifierJob> jobqueue;
		List<Thread> threads;
		int numThreads;
		public StreamWriter jobwriter;
		HashSet<int> trueRows;

		public override void loadArguments(string[] args)
		{
			generator = new CandidateFileGenerator(datadir+args[0]);
			datafile = datadir+args[1];

			List<int> mnp = new List<int>();
			List<int> bw = new List<int>();
			List<int> mpl = new List<int>();
			List<double> alphas = new List<double>();

			numThreads = int.Parse(args[2]);

			string[] parts = args[3].Split(' ');
			foreach (string s in parts)
				mnp.Add(int.Parse(s));

			parts = args[4].Split(' ');
			foreach (string s in parts)
				bw.Add(int.Parse(s));

			parts = args[5].Split(' ');
			foreach (string s in parts)
				mpl.Add(int.Parse(s));

			parts = args[6].Split(' ');
			foreach (string s in parts)
				alphas.Add(double.Parse(s, CultureInfo.InvariantCulture));

			jobqueue = new ThreadSafeQueue<TestClassifierJob>();
			foreach (int mn in mnp)
			{
				foreach (int b in bw)
				{
					foreach (int mp in mpl)
					{
						foreach (double a in alphas)
						{
							TestClassifierJob job = new TestClassifierJob(mn, b, mp, a);
							jobqueue.Enqueue(job);
						}
					}
				}
			}
		}

		public override void start()
		{
			jobwriter = new StreamWriter(expdir + expName + ".csv");
			String s = ";";
			String line = "Maximum number of patterns" + s + "Maximum pattern length" + s + "Beam width" + s + "alpha";
			line += s + "Duration (s)" + s + "Number of patterns" + s + "TP" + s + "FP" + s + "TN" + s + "FN";
			line += s + "Precision" + s + "Recall"+s+"g-Measure";
			explog(line);

			log("Candidatesfile: " + datafile);

			Console.Write("Reading description data.. ");
			matrix = new BinaryMatrixBitArrays(datafile);
			Console.WriteLine("Done!");

			log("Description data: " + matrix.M + "x" + matrix.N);
			Console.Write("Creating community");
			ThreadSafeQueue<Community> queue = generator.generate(graph,matrix);
			Community c = queue.Dequeue();
			trueRows = new HashSet<int>(c.getVertices().Keys);
			Console.WriteLine("Done! Community: " + c.Size);
			log("Community size: " + c.Size);

			Console.Write("Creating threads.. ");
			threads = new List<Thread>();
			for (int i = 0; i < numThreads; i++)
			{
				Thread t = new Thread(this.run);
				threads.Add(t);
				t.Start();
			}
			log("Num threads: " + numThreads);
			Console.WriteLine("Done!");
			
			foreach (Thread t in threads)
				t.Join();
			
			jobwriter.Close();
		}


		public void run()
		{
			while (!jobqueue.Empty())
			{
				try
				{
					int threadId = 0;
					foreach (Thread t in threads)
					{
						if (t == Thread.CurrentThread)
							break;
						threadId++;
					}

					TestClassifierJob job = jobqueue.Dequeue();
					Console.WriteLine("Thread " + threadId + " starts on maxNumPatterns:" + job.maxNumPatterns + ", maxPatternLength:" + job.maxPatternLength + ", beam:" + job.beamWidth+", alpha: "+job.alpha);
					TimeMeasure measure = new TimeMeasure();
					DisjunctionBuilder builder = new DisjunctionBuilder(matrix, job.maxNumPatterns, job.beamWidth, job.maxPatternLength, job.alpha);
					DisjunctionClassifier classifier = (DisjunctionClassifier)builder.build(new HashSet<int>(trueRows));
					double duration = measure.getDuration();
					int[] q = classifier.quality(matrix, trueRows);
					String s = ";";
					String line = job.maxNumPatterns+s+job.maxPatternLength+s+job.beamWidth+s+job.alpha;
					line+= s+duration+s+classifier.patterns.Count+s+q[0]+s+q[1]+s+q[2]+s+q[3];
					line += s + classifier.precision(q) + s + classifier.recall(q) + s+builder.measure(q[0], q[1], q[2], q[3]);
					explog(line);
					Console.WriteLine("Thread " + threadId + " finished in " + duration + " seconds");
					Console.WriteLine("Queue size: " + jobqueue.Count);
				}
				catch (Exception)
				{
					return;
				}
			}
		}
	
		public void explog(string msg)
		{
			lock(jobwriter)
			{
				jobwriter.WriteLine(msg);
				jobwriter.Flush();
			}
		}
	}
}
